"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Compile = void 0;
const grammar_1 = require("./grammar");
const ops_1 = __importDefault(require("../libs/ops"));
const stdlib_1 = __importDefault(require("../libs/stdlib"));
const simplify_1 = require("./simplify");
const types_1 = require("../types");
const path_1 = __importDefault(require("path"));
const util_1 = require("./util");
const piecewiseOps_1 = __importDefault(require("../libs/piecewiseOps"));
const templates_1 = require("./templates");
/**
 * Compiles LogiMat to a math function (or multiple). Each function/variable will be on a separate line.
 * @param input {string} - is the input LogiMat code that will be compiled.
 * @param useTex {boolean} - is a value indicating if the output should be converted to Tex.
 * @param noFS {boolean} - is a valid indicating if untrusted filesystem operations should be blocked (for example, code telling the compiler to load an NPM module). This is not a security feature.
 * @param filePath {string} - is a path to the file currently being compiled.
 * @param piecewise {boolean} - is a value indicating if the output should use piecewise instead of pure math for logic.
 * @param strict {boolean} - is a value indicating if an error should be thrown if an undefined function/variable is used.
 * @param outputMaps {boolean} - is a value indicating if the output should be a map of simplified to unsimplified output.
 * @param simplificationMap {Record<string, string>} - is a map of simplified values to unsimplified values.
 * @param importMap {Record<string, TemplateModule>} - is a map of imports which will be used instead of require if specified.
 */
const Compile = (input, useTex = false, noFS = false, filePath = null, piecewise = false, strict = false, outputMaps = false, simplificationMap = {}, importMap = {}) => __awaiter(void 0, void 0, void 0, function* () {
    const tree = GetTree(input);
    const state = { logimat: { files: [], definitions: {} } };
    if (filePath)
        state.logimat.files = [filePath];
    if (!noFS && filePath) {
        module.paths.push(path_1.default.dirname(filePath));
    }
    const templates = (0, templates_1.createTemplates)(noFS, state, importMap);
    const postTemplates = [];
    for (const declaration of tree.imports) {
        if (noFS)
            throw new Error("Import failed: filesystem operations have been disabled.");
        if (declaration.importType === "template") {
            try {
                const module = importMap.hasOwnProperty(declaration.path) ? importMap[declaration.path] : require(declaration.path);
                if (!(module === null || module === void 0 ? void 0 : module.hasOwnProperty("templates")))
                    throw new Error("Module does not contain templates!");
                if (typeof (module.templates) !== "object" || Array.isArray(module.templates))
                    throw new Error("Templates is not an object!");
                for (const templateKey of Object.keys(module.templates)) {
                    const template = module.templates[templateKey];
                    if (typeof (template) !== "object" || Array.isArray(module.templates) || !template.hasOwnProperty("function") || typeof (template.function) !== "function") {
                        throw new Error("Template \"" + templateKey + "\" on module \"" + declaration.path + "\" is not defined correctly!");
                    }
                    templates[templateKey.trim().toLowerCase()] = template.function;
                }
                if (!module.hasOwnProperty("postTemplates"))
                    continue;
                if (typeof (module.postTemplates) !== "string")
                    throw new Error("PostTemplates is not a string!");
                postTemplates.push(module.postTemplates);
            }
            catch (e) {
                console.error("An error occurred while loading module \"" + declaration.path + "\".");
                throw e;
            }
        }
    }
    const templatesRef = {
        handledTemplates: true
    };
    let count = 0;
    let declarations;
    try {
        declarations = yield TraverseTemplatesArr(tree.declarations, templates, state, templatesRef, simplificationMap);
        //Allow up to 50 layers of functions.
        while (templatesRef.handledTemplates && count < 50) {
            templatesRef.handledTemplates = false;
            declarations = yield TraverseTemplatesArr(declarations, templates, state, templatesRef, simplificationMap);
            count++;
        }
        declarations.push(...GetTree(postTemplates.join("\n")).declarations);
        //Reset variables.
        templatesRef.handledTemplates = true;
        count = 0;
        //Allow up to 50 layers of functions.
        while (templatesRef.handledTemplates && count < 50) {
            templatesRef.handledTemplates = false;
            declarations = yield TraverseTemplatesArr(declarations, templates, state, templatesRef, simplificationMap);
            count++;
        }
    }
    catch (e) {
        //Log the file the error occurred in, if it exists.
        if (state.logimat.files.length > 0)
            console.error("An error occurred in \"" + state.logimat.files[state.logimat.files.length - 1] + "\":");
        throw e;
    }
    // Next, we want to handle export declarations. Exports work by creating an export with a random name, along with the contents
    // of the inline specified by the export. Next, the original inline's contents are changed to point to the new export.
    const fileInlines = GetInlines(declarations);
    const stdLibInlines = GetInlines(GetTree(stdlib_1.default).declarations);
    const opsInlines = GetInlines(GetTree((piecewise ? piecewiseOps_1.default : ops_1.default)).declarations);
    const exportInlines = Object.assign(Object.assign(Object.assign({}, fileInlines), stdLibInlines), opsInlines);
    const exportDeclarations = [];
    let varIdx = 1;
    for (const exportName of GetExportDeclarations(declarations)) {
        // This exportName is the name of an inline.
        // First, we should look up the inline.
        const inline = exportInlines[exportName];
        // If it doesn't exist, no need to proceed.
        if (!inline)
            continue;
        // Now, let's create a variable for the new export.
        const name = "v_" + varIdx++;
        // Remove the value from declarations.
        const idx = declarations.indexOf(inline.value);
        if (idx !== -1)
            declarations.splice(idx, 1);
        if (inline.function) {
            // Create an export function.
            const exportFunction = inline.value;
            const inlineName = exportFunction.name;
            exportFunction.name = name;
            exportFunction.modifier = "export";
            // Create args which are the length of the original's args.
            const inlineArgs = [...exportFunction.args];
            exportFunction.args = Array(inlineArgs.length).fill("").map(() => "v_" + varIdx++);
            // Finally, we to insert something to map between the old and new names.
            const maps = inlineArgs.map((inlineArg, idx) => {
                const exportArg = exportFunction.args[idx];
                return {
                    type: "const",
                    name: inlineArg,
                    expr: {
                        type: "v",
                        args: [exportArg]
                    }
                };
            });
            exportFunction.block.unshift(...maps);
            // Create an inline function.
            const inlineFunction = {
                type: "function",
                modifier: "inline",
                name: inlineName,
                args: inlineArgs,
                block: [
                    {
                        type: "state",
                        expr: {
                            type: "f",
                            args: [
                                name,
                                inlineArgs.map(arg => ({
                                    type: "v",
                                    args: [
                                        arg
                                    ]
                                }))
                            ]
                        }
                    }
                ]
            };
            // Add the new functions to declarations.
            declarations.push(exportFunction);
            exportDeclarations.push(inlineFunction);
        }
        else {
            // Create an export const.
            const exportConst = inline.value;
            const inlineName = exportConst.name;
            exportConst.name = name;
            exportConst.modifier = "export";
            // Create an inline const.
            const inlineConst = {
                type: "const",
                modifier: "inline",
                name: inlineName,
                expr: {
                    type: "v",
                    args: [
                        name
                    ]
                }
            };
            // Add the new functions to declarations.
            declarations.push(exportConst);
            exportDeclarations.push(inlineConst);
        }
    }
    // We need to run GetInlines again because declarations is changed.
    const inlines = Object.assign(Object.assign(Object.assign(Object.assign({}, GetInlines(declarations)), stdLibInlines), opsInlines), GetInlines(exportDeclarations));
    const stack = [];
    try {
        const output = InternalCompile(useTex, declarations, inlines, templates, state, varIdx, stack, strict, simplificationMap);
        if (outputMaps)
            return Object.assign(Object.assign({}, output), { importMap });
        else
            return output.output.join("\n");
    }
    catch (e) {
        if (stack.length > 0)
            console.error("Call stack: " + stack.slice(0, Math.min(20, stack.length)).join(" -> "));
        throw e;
    }
});
exports.Compile = Compile;
const TraverseTemplatesArr = (input, templates, state, ref, simplificationMap) => __awaiter(void 0, void 0, void 0, function* () {
    const output = [];
    for (const val of input) {
        if (Array.isArray(val)) {
            output.push([...yield TraverseTemplatesArr(val, templates, state, ref, simplificationMap)]);
        }
        else if (typeof (val) === "object") {
            const newVal = yield TraverseTemplatesObj(val, templates, state, ref, simplificationMap);
            if (Array.isArray(newVal))
                output.push(...newVal);
            else
                output.push(newVal);
        }
        else {
            output.push(val);
        }
    }
    return output;
});
const TraverseTemplatesObj = (input, templates, state, ref, simplificationMap) => __awaiter(void 0, void 0, void 0, function* () {
    if (input == null)
        return null;
    if (input.hasOwnProperty("type") && ["template", "templatefunction"].includes(input["type"])) {
        return yield HandleTemplate(input, templates, state, ref, simplificationMap);
    }
    const output = {};
    for (const key in input) {
        const val = input[key];
        if (Array.isArray(val)) {
            output[key] = yield TraverseTemplatesArr(val, templates, state, ref, simplificationMap);
        }
        else if (typeof (val) === "object") {
            output[key] = yield TraverseTemplatesObj(val, templates, state, ref, simplificationMap);
        }
        else {
            output[key] = val;
        }
    }
    return output;
});
const HandleTemplate = (templateDeclaration, templates, state, ref, simplificationMap) => __awaiter(void 0, void 0, void 0, function* () {
    var _a, _b;
    let output;
    const name = (_b = (_a = templateDeclaration.name) === null || _a === void 0 ? void 0 : _a.trim()) === null || _b === void 0 ? void 0 : _b.toLowerCase();
    if (templateDeclaration.type === "templatefunction") {
        // @ts-ignore
        output = templateDeclaration.function(state);
        ref.handledTemplates = true;
    }
    else {
        if (!templates.hasOwnProperty(name))
            throw new Error("Template \"" + templateDeclaration.name + "\" does not exist!");
        try {
            //Handle expressions in the template args.
            const templateArgs = [];
            for (const arg of templateDeclaration.args) {
                if (arg && arg["expression"]) {
                    const definedNames = Object.keys(state.logimat.definitions).filter(key => typeof (state.logimat.definitions[key]) === "number");
                    const handled = yield TraverseTemplatesObj(arg["value"], templates, state, ref, simplificationMap);
                    //Handle special cases like raw numbers and non-numerical definitions.
                    if (handled.hasOwnProperty("type")) {
                        if (handled["type"] === "v" && handled["args"] && state.logimat.definitions.hasOwnProperty(handled["args"][0])) {
                            templateArgs.push(state.logimat.definitions[handled["args"][0]]);
                            continue;
                        }
                    }
                    else {
                        const value = Object.values(handled).join("");
                        if (!(0, util_1.isNumeric)(value)) {
                            if (arg["nonStrict"] && typeof (handled) === "string") {
                                templateArgs.push(handled);
                                continue;
                            }
                            throw new Error("The input \"" + arg["source"] + "\" cannot be evaluated to a number. Expressions input into templates must evaluate to numbers, and can only use defined variables.");
                        }
                        templateArgs.push(parseFloat(value));
                        continue;
                    }
                    const compiled = CompileExpression(handled, {
                        inlines: {},
                        varIdx: { value: 0 },
                        names: definedNames,
                        stack: [],
                        state,
                        templates,
                        vars: Object.fromEntries(Object.entries(state.logimat.definitions).filter(([key, value]) => typeof (value) === "number"))
                    });
                    const simplified = (0, simplify_1.SimplifyExpression)(compiled, false, !arg["nonStrict"], definedNames, simplificationMap);
                    if (!(0, util_1.isNumeric)(simplified)) {
                        if (arg["nonStrict"]) {
                            templateArgs.push(simplified);
                            continue;
                        }
                        throw new Error("The input \"" + arg["source"] + "\" cannot be evaluated to a number. Expressions input into templates must evaluate to numbers, and can only use defined variables.");
                    }
                    templateArgs.push(parseFloat(simplified));
                    continue;
                }
                else if (typeof (arg) === "object") {
                    templateArgs.push(yield TraverseTemplatesObj(arg, templates, state, ref, simplificationMap));
                    continue;
                }
                templateArgs.push(arg);
            }
            output = yield templates[name](templateArgs, state, templateDeclaration.context);
        }
        catch (e) {
            console.error("An error occurred while running the \"" + templateDeclaration.name + "\" template:");
            throw e;
        }
        //Ignore the setFile template. It's an internal template that returns itself in order to maintain file boundaries, but should not keep
        //template resolution from exiting. What this does is instead of handling the templates of the output, we simply return the template.
        //Doing this means that it will only be read until the next iteration.
        if (["setfile"].includes(name)) {
            return templateDeclaration;
        }
        ref.handledTemplates = true;
    }
    if (output instanceof Function) {
        return { type: "templatefunction", context: templateDeclaration.context, function: output };
    }
    if (["concat", "parse", "wrap"].includes(name))
        return output;
    //TODO: Allow templates to import other templates.
    switch (templateDeclaration.context) {
        case types_1.TemplateContext.OuterDeclaration:
            return yield TraverseTemplatesArr(GetTree(output).declarations, templates, state, ref, simplificationMap);
        case types_1.TemplateContext.InnerDeclaration:
            return yield TraverseTemplatesArr(GetStatementsTree(output), templates, state, ref, simplificationMap);
        case types_1.TemplateContext.Expression:
            const expr = GetExpression(output);
            if (typeof (expr) === "object") {
                return yield TraverseTemplatesObj(expr, templates, state, ref, simplificationMap);
            }
            return expr;
    }
    return [];
});
const InternalCompile = (useTex, tree, inlines, templates, state, varIdx, stack, strict, simplificationMap = {}) => {
    const out = [];
    const outerNames = GetDeclaredNames(tree);
    const display = {};
    let data = { inlines, templates, state, stack, names: [], vars: {}, varIdx: { value: varIdx } };
    for (const declaration of tree) {
        if (declaration.modifier === "inline")
            continue;
        const names = Object.assign([], outerNames);
        data = {
            inlines,
            templates,
            state,
            stack,
            names,
            vars: {},
            varIdx: data.varIdx
        };
        switch (declaration.type) {
            case "display":
                let val;
                if (typeof (declaration.value) === "string")
                    val = declaration.value;
                else if (declaration.value.type === "tstring")
                    val = declaration.value.args.map(arg => {
                        if (typeof (arg) === "string")
                            return arg;
                        else
                            return "${" + (0, simplify_1.SimplifyExpression)(CompileExpression(arg, data), useTex, strict, names, simplificationMap) + "}";
                    }).join("");
                else if (declaration.value.type === "aargs")
                    val = (0, util_1.HandleName)(declaration.value.name) + "(" + declaration.value.args.map(arg => {
                        if (typeof (arg) === "string") {
                            if (arg === "index")
                                return "\\operatorname{index}";
                            return arg;
                        }
                        else
                            return (0, simplify_1.SimplifyExpression)(CompileExpression(arg, data), useTex, strict, names, simplificationMap);
                    }).join(",") + ")";
                else
                    val = (0, simplify_1.SimplifyExpression)(CompileExpression(declaration.value, data), useTex, strict, names, simplificationMap);
                display[declaration.displayType] = val;
                break;
            case "function":
                out.push(...CompileDisplay(display));
                const functionDeclaration = declaration;
                out.push((0, util_1.HandleName)(functionDeclaration.name) + "(" + functionDeclaration.args.map(util_1.HandleName).join(",") + ")" + "=" + (0, simplify_1.SimplifyExpression)(CompileBlock(functionDeclaration.block, data, "", {}), useTex, strict, names.concat(functionDeclaration.args), simplificationMap));
                break;
            case "const":
                out.push(...CompileDisplay(display));
                const constDeclaration = declaration;
                out.push((0, util_1.HandleName)(constDeclaration.name) + "=" + (0, simplify_1.SimplifyExpression)(CompileExpression(constDeclaration.expr, data), useTex, strict, names, simplificationMap));
                break;
            case "action":
                const actionDeclaration = declaration;
                out.push((actionDeclaration.funcName ? (0, util_1.HandleName)(actionDeclaration.funcName) + (actionDeclaration.args ? "(" + actionDeclaration.args.map(util_1.HandleName).join(",") + ")" : "") + "=" : "") + (0, util_1.HandleName)(actionDeclaration.name) + "\\to " + (0, simplify_1.SimplifyExpression)(CompileBlock(actionDeclaration.block, data, "", {}), useTex, strict, actionDeclaration.args ? names.concat(actionDeclaration.args) : names, simplificationMap));
                break;
            case "actions":
                const actionsDeclaration = declaration;
                //Get the name without the args (the first part before the opening parenthesis)
                const badActions = actionsDeclaration.args.filter(action => !names.includes(action[0]));
                if (strict && badActions.length > 0) {
                    throw new Error("The following actions have not been defined: " + badActions.map(action => "\"" + action[0] + "\"").join(", ") + ".");
                }
                out.push((0, util_1.HandleName)(actionsDeclaration.name) + (actionsDeclaration.actionArgs ? "(" + actionsDeclaration.actionArgs.map(util_1.HandleName).join(",") + ")" : "") + "=" + actionsDeclaration.args.map(action => (0, util_1.HandleName)(action[0]) + (action.length > 1 ? "(" + action.slice(1).map(util_1.HandleName) + ")" : "")).join(","));
                break;
            case "expression":
                out.push(...CompileDisplay(display));
                const expressionDeclaration = declaration;
                out.push((0, simplify_1.SimplifyExpression)(CompileBlock(expressionDeclaration.block, data, "", {}), useTex, strict, names, simplificationMap));
                break;
            case "graph":
                out.push(...CompileDisplay(display));
                //Give the graph access to x and y.
                names.push("x", "y");
                const graphDeclaration = declaration;
                out.push((0, simplify_1.SimplifyExpression)(CompileExpression(graphDeclaration.p1, data), useTex, strict, names, simplificationMap) + util_1.opMap[graphDeclaration.op] + (0, simplify_1.SimplifyExpression)(CompileExpression(graphDeclaration.p2, data), useTex, strict, names, simplificationMap));
                break;
            case "point":
                out.push(...CompileDisplay(display));
                const pointDeclaration = declaration;
                out.push((0, simplify_1.SimplifyExpression)(CompileExpression(pointDeclaration.point, data), useTex, strict, names, simplificationMap));
                break;
            case "polygon":
                out.push(...CompileDisplay(display));
                const polygonDeclaration = declaration;
                out.push("\\operatorname{polygon}" + (useTex ? "\\left(" : "(") + polygonDeclaration.points.map(point => (0, simplify_1.SimplifyExpression)(CompileExpression(point, data), useTex, strict, names, simplificationMap)).join(",") + (useTex ? "\\right)" : ")"));
                break;
        }
    }
    return { output: out, simplificationMap };
};
const CompileDisplay = (input) => {
    const out = Object.entries(input).map(entry => "!" + entry[0] + "=" + entry[1]);
    //Clear the object.
    for (const key in input) {
        delete input[key];
    }
    return out;
};
const GetTree = (input) => {
    const match = grammar_1.grammar.match(input);
    if (match.failed())
        throw new Error(match.message);
    return (0, grammar_1.semantic)(match).parse();
};
const GetStatementsTree = (input) => {
    if (input.trim() === "")
        return [];
    const match = grammar_1.grammar.match(input, "InnerDeclarations");
    if (match.failed())
        throw new Error(match.message);
    return (0, grammar_1.semantic)(match).parse();
};
const GetExpression = (input) => {
    const match = grammar_1.grammar.match(input, "Expression");
    if (match.failed())
        throw new Error(match.message);
    return (0, grammar_1.semantic)(match).parse();
};
const GetInlines = (tree) => {
    const inlines = {};
    for (const declaration of tree) {
        if (declaration.modifier === "inline")
            inlines[declaration["name"]] = { function: declaration.type === "function", value: declaration };
    }
    return inlines;
};
const GetExportDeclarations = (tree) => {
    const exports = [];
    for (const declaration of tree) {
        if (declaration.type === "export" && !exports.includes(declaration.name))
            exports.push(declaration.name);
    }
    return exports;
};
const GetDeclaredNames = (tree) => {
    const output = [];
    for (const declaration of tree) {
        if (declaration.type === "action") {
            if (declaration["funcName"])
                output.push(declaration["funcName"]);
        }
        else {
            if (declaration["name"])
                output.push(declaration["name"]);
        }
    }
    return output;
};
const CompileBlock = (input, data, defaultOut, args) => {
    let out = defaultOut;
    let newVars = Object.assign(Object.assign({}, data.vars), args);
    for (const statement of input) {
        switch (statement.type) {
            case "const":
                newVars[statement["name"]] = CompileExpression(statement["expr"], Object.assign(Object.assign({}, data), { vars: Object.assign(Object.assign({}, newVars), { state: out || data.vars["state"] || "" }) }));
                break;
            case "state":
                out = CompileExpression(statement["expr"], Object.assign(Object.assign({}, data), { vars: Object.assign(Object.assign({}, newVars), { state: out || data.vars["state"] || "" }) }));
                break;
            case "if":
                if (!statement["elseaction"] && !out) {
                    throw new Error("The state must be set before an else-less if can be used.");
                }
                const condition = CompileExpression(statement["condition"], Object.assign(Object.assign({}, data), { vars: Object.assign(Object.assign({}, newVars), { state: out }) }));
                const ifaction = CompileBlock(statement["ifaction"], Object.assign(Object.assign({}, data), { vars: Object.assign(Object.assign({}, newVars), { state: out }) }), out, {});
                const elseaction = statement["elseaction"] ? CompileBlock(statement["elseaction"], Object.assign(Object.assign({}, data), { vars: Object.assign(Object.assign({}, newVars), { state: out }) }), out, {}) : out;
                out = CompileExpression({ type: "f", args: ["if_func", [condition, ifaction, elseaction]] }, Object.assign(Object.assign({}, data), { vars: Object.assign(Object.assign({}, newVars), { state: out || data.vars["state"] || "" }) }));
                break;
        }
    }
    if (out === "")
        throw new Error("The state must be set inside of every block.");
    return out;
};
/**
 * Handles a logic function and transforms it to a string, for when piecewise output is enabled.
 * @param name {string} - is the name of the function.
 * @param args {(string | object)[]} - is the array of arguments to the function.
 * @returns {string} The resultant string.
 */
const HandlePiecewise = (name, args) => {
    switch (name) {
        case "and":
            return `((${args[0]}) & (${args[1]}))`;
        case "or":
            return `((${args[0]}) | (${args[1]}))`;
        case "equal":
            return `((${args[0]}) == (${args[1]}))`;
        case "notEqual":
            return `if_func((${args[0]}) == (${args[1]}), 0, 1)`;
        case "lt":
            return `((${args[0]}) < (${args[1]}))`;
        case "lte":
            return `((${args[0]}) <= (${args[1]}))`;
        case "gt":
            return `((${args[0]}) > (${args[1]}))`;
        case "gte":
            return `((${args[0]}) >= (${args[1]}))`;
        default:
            return "";
    }
};
const piecewiseFunctions = [
    "and",
    "or",
    "equal",
    "notEqual",
    "lt",
    "lte",
    "gt",
    "gte"
];
const CompileExpression = (expression, data) => {
    if (typeof (expression) !== "object")
        return expression;
    const args = expression.args.map(arg => typeof (arg) === "object" && arg.hasOwnProperty("type") && arg.hasOwnProperty("args") ? CompileExpression(arg, data) : arg);
    switch (expression.type) {
        case "+":
            return "(" + args[0] + ")+(" + args[1] + ")";
        case "-":
            return "(" + args[0] + ")-(" + args[1] + ")";
        case "*":
            return "(" + args[0] + ")*(" + args[1] + ")";
        case "/":
            return "(" + args[0] + ")/(" + args[1] + ")";
        case "^":
            return "pow(" + args[0] + "," + args[1] + ")";
        case "n":
            return "-(" + args[0] + ")";
        case "f":
            const fargs = expression.args[1].map(arg => typeof (arg) === "object" && arg.hasOwnProperty("type") ? CompileExpression(arg, data) : arg);
            if (!data.inlines.hasOwnProperty(expression.args[0])) {
                if (piecewiseFunctions.includes(expression.args[0]))
                    return HandlePiecewise(expression.args[0], fargs);
                return expression.args[0] + "(" + fargs.join(",") + ")";
            }
            const fargnames = data.inlines[expression.args[0]].value.args;
            if (fargnames.length !== fargs.length)
                throw new Error("Inline function \"" + expression.args[0] + "\" requires " + fargnames.length + ", but only " + fargs.length + " are given.");
            data.stack.push(expression.args[0]);
            const result = CompileBlock(data.inlines[expression.args[0]].value.block, data, "", Object.fromEntries(fargnames.map((v, i) => [v, fargs[i]])));
            data.stack.pop();
            return "(" + result + ")";
        case "v":
            const name = expression.args[0];
            if (data.vars.hasOwnProperty(name))
                return data.vars[name];
            if (data.inlines.hasOwnProperty(name))
                return CompileExpression(data.inlines[name].value["expr"], data);
            switch (name) {
                case "pi":
                    return "pi()";
                case "infinity":
                case "infty":
                case "inf":
                    return "inf()";
            }
            return name;
        case "sum":
            const sumName = "v_" + data.varIdx.value++;
            //Map the user-chosen variable to generated name.
            const sumVar = { [args[0]]: sumName };
            //Make the variable name used here a declared variable, in order to make it work in strict mode.
            data.names.push(sumName);
            return "sum(" + sumName + "," + args[1] + "," + args[2] + "," + CompileBlock(args[3], Object.assign(Object.assign({}, data), { vars: Object.assign(Object.assign({}, data.vars), sumVar) }), "", {}) + ")";
        case "prod":
            const prodName = "v_" + data.varIdx.value++;
            //Map the user-chosen variable to generated name.
            const prodVar = { [args[0]]: prodName };
            //Make the variable name used here a declared variable, in order to make it work in strict mode.
            data.names.push(prodName);
            return "prod(" + prodName + "," + args[1] + "," + args[2] + "," + CompileBlock(args[3], Object.assign(Object.assign({}, data), { vars: Object.assign(Object.assign({}, data.vars), prodVar) }), "", {}) + ")";
        case "int":
            const intName = "v_" + data.varIdx.value++;
            //Map the user-chosen variable to generated name.
            const intVar = { [args[0]]: intName };
            //Make the variable name used here a declared variable, in order to make it work in strict mode.
            data.names.push(intName);
            return "int(" + intName + "," + args[1] + "," + args[2] + "," + CompileBlock(args[3], Object.assign(Object.assign({}, data), { vars: Object.assign(Object.assign({}, data.vars), intVar) }), "", {}) + ")";
        case "div":
            return "div(" + args[0] + "," + CompileBlock(args[1], Object.assign(Object.assign({}, data), { vars: Object.assign({}, data.vars) }), "", {}) + ")";
        case "b":
            return CompileBlock(expression.args[0], data, "", {});
        case "a_f":
            //Map the user-chosen variable to the array for Desmos' filter syntax.
            const filterVar = { [args[1]]: args[0] };
            //If it's an array (list of statements), then compile it as a block, otherwise compile the original as an expression (with the .
            const filterFunc = Array.isArray(args[2])
                ? CompileBlock(args[2], Object.assign(Object.assign({}, data), { vars: Object.assign(Object.assign({}, data.vars), filterVar) }), "", {})
                : CompileExpression(expression.args[2], Object.assign(Object.assign({}, data), { vars: Object.assign(Object.assign({}, data.vars), filterVar) }));
            return "array_filter(" + args[0] + "," + filterFunc + ")";
        case "a_m":
            const mapName = "v_" + data.varIdx.value++;
            //Map the user-chosen variable to generated name.
            const mapVar = { [args[1]]: mapName };
            //Make the variable name used for mapping a declared variable, in order to make it work in strict mode.
            data.names.push(mapName);
            //If it's an array (list of statements), then compile it as a block, otherwise compile the original as an expression.
            const mapFunc = Array.isArray(args[2])
                ? CompileBlock(args[2], Object.assign(Object.assign({}, data), { vars: Object.assign(Object.assign({}, data.vars), mapVar) }), "", {})
                : CompileExpression(expression.args[2], Object.assign(Object.assign({}, data), { vars: Object.assign(Object.assign({}, data.vars), mapVar) }));
            return "array_map(" + args[0] + "," + mapFunc + "," + mapName + ")";
    }
    return "";
};
//# sourceMappingURL=compiler.js.map